@@ -9,7 +9,7 @@ USER_OBJECT_
 RS_GGOBI(getDisplayWindow)(USER_OBJECT_ display)
 {
     displayd *dpy;
-    dpy = (displayd*) (long) NUMERIC_DATA(VECTOR_ELT(display, 1))[0];
+    dpy = (displayd*) R_ExternalPtrAddr((VECTOR_ELT(display, 1)));
     
 
     if(!dpy || GTK_IS_GGOBI_WINDOW_DISPLAY(dpy) == false)
@@ -32,8 +32,8 @@ RS_GGOBI(getDisplayPlotWidgets)(USER_OBJECT_ display)
    GList *tmp;
    USER_OBJECT_ els;
 
-#if 0
-    dpy =  R_ExternalPtrAddr(display);
+#if 1
+    dpy =  R_ExternalPtrAddr(VECTOR_ELT(display, 1));
 #else
     dpy = (displayd*) (long) NUMERIC_DATA(VECTOR_ELT(display, 1))[0];
 #endif
