@@ -32,7 +32,7 @@ RS_GGOBI(setBrushSize)(USER_OBJECT_ dims, USER_OBJECT_ datasetId, USER_OBJECT_ g
      INTEGER_DATA(ans)[1] = ht;
 
    /* experiments */
-   brush_reset(gg,RESET_UNHIDE_POINTS);
+   brush_reset(gg,RESET_UNSHADOW_POINTS);
    /*   displays_plot (NULL, FULL, gg); */
      gdk_flush();
   } else
