@@ -10,10 +10,14 @@
 
 
 #ifdef G_OS_WIN32
+#if 0
 #ifdef ERROR
 #undef ERROR
 #endif
-#include <gdk/win32/gdkwin32.h>
+#endif
+#if 0
+#include <gdk/gdkwin32.h>
+#endif
 #include <process.h>
 static void __cdecl GtkEventThreadHandler(void *display);
 void R_gtk_handle_events();
@@ -47,20 +51,18 @@ USER_OBJECT_ RS_INTERNAL_GGOBI(describeDataset)(datad *d, gboolean full);
 USER_OBJECT_ RS_GGOBI(getDisplays)(USER_OBJECT_ rsFull, USER_OBJECT_ ggobiId);
 USER_OBJECT_ RS_INTERNAL_GGOBI(getDisplays)(gboolean full, ggobid *gg);
 
+
 /*
   Create a ggobi instance and initialize.
  */
 USER_OBJECT_
 RS_GGOBI(init)(USER_OBJECT_ args, USER_OBJECT_ createInstance)
 {
-
  gint which, i;
  char **c_args;
  int n = GET_LENGTH(args);
  USER_OBJECT_ ans;
 
-   GGobi_setMissingValueIdentifier(R_IsNaNorNA);
-
    c_args = g_malloc(sizeof(char *)*n);
    for(i = 0; i < n ; i++) {
      c_args[i] = CHAR_DEREF(STRING_ELT(args, i));
@@ -76,7 +78,7 @@ RS_GGOBI(init)(USER_OBJECT_ args, USER_OBJECT_ createInstance)
    g_free(c_args);
 
 #ifdef G_OS_WIN32 
-  R_tcldo = R_gtk_handle_events;
+   /*XXX R_tcldo = R_gtk_handle_events; */
 #else
   addInputHandler (R_InputHandlers, ConnectionNumber(gdk_display),
                    RS_GGOBI(event_handle), -1);
@@ -210,7 +212,10 @@ RS_INTERNAL_GGOBI(describeDataset)(datad *d, gboolean full)
  SET_VECTOR_ELT(ans, 2, tmp = NEW_INTEGER(1));
   INTEGER_DATA(tmp)[0] = d->input->mode;
   SET_NAMES(tmp, tmp1 = NEW_CHARACTER(1));
+#if 0
+/*XXX */
   SET_STRING_ELT(tmp1, 0, COPY_TO_USER_STRING(GGOBI(getDataModeDescription)(INTEGER_DATA(tmp)[0])));
+#endif
   SET_NAMES(tmp, tmp1);
   SET_STRING_ELT(names, 2, COPY_TO_USER_STRING("format"));
 
@@ -294,6 +299,9 @@ RS_GGOBI(getDataModes)()
     SET_STRING_ELT(names, i, COPY_TO_USER_STRING(modeNames[i]));
   }
 
+  if(modeNames)
+     g_free((gchar *)modeNames);
+
   SET_NAMES(ans, names);
   UNPROTECT(2);
   return(ans);
@@ -310,13 +318,18 @@ RS_GGOBI(setFile)(USER_OBJECT_ fileName, USER_OBJECT_ smode, USER_OBJECT_ add, U
  DataMode mode;
  ggobid *gg = GetGGobi(gobiId);
  USER_OBJECT_ ans = NEW_INTEGER(1);
+ gchar *modeName;
 
+ if(IS_CHARACTER(smode))
+	 modeName = CHAR_DEREF(STRING_ELT(smode, 0));
+ else if(IS_INTEGER(smode)) {
   mode = INTEGER_DATA(smode)[0];
   if(mode < 0)
     mode = unknown_data;
+ }
 
   /*if(fileset_read_init(CHAR_DEREF(STRING_ELT(fileName, 0)), mode, gg)) {*/
-  if(fileset_read_init(CHAR_DEREF(STRING_ELT(fileName, 0)), gg)) {
+  if(fileset_read_init(CHAR_DEREF(STRING_ELT(fileName, 0)), modeName, NULL, gg)) {
     INTEGER_DATA(ans)[0] = g_slist_length(gg->d)-1;
     display_menu_build(gg);
   } else
@@ -492,10 +505,10 @@ RS_GGOBI(getDisplayOptions)(USER_OBJECT_ which, USER_OBJECT_ ggobiId)
 
   LOGICAL_DATA(ans)[DOPT_AXESLAB] = options->axes_label_p;
   SET_STRING_ELT(names, DOPT_AXESLAB,
-    COPY_TO_USER_STRING("Show 2D tour axes as text"));
+    COPY_TO_USER_STRING("Show tour axes"));
   LOGICAL_DATA(ans)[DOPT_AXESVALS] = options->axes_values_p;
   SET_STRING_ELT(names, DOPT_AXESVALS,
-    COPY_TO_USER_STRING("Show 2D tour axes as text"));
+    COPY_TO_USER_STRING("Show axes labels"));
 
   LOGICAL_DATA(ans)[DOPT_EDGES_U] = options->edges_undirected_show_p;
   SET_STRING_ELT(names, DOPT_EDGES_U, COPY_TO_USER_STRING("Undirected edges"));
@@ -506,7 +519,7 @@ RS_GGOBI(getDisplayOptions)(USER_OBJECT_ which, USER_OBJECT_ ggobiId)
 
   LOGICAL_DATA(ans)[DOPT_WHISKERS] = options->whiskers_show_p;
   SET_STRING_ELT(names, DOPT_WHISKERS,
-    COPY_TO_USER_STRING("Segments for PC and TS plots"));
+    COPY_TO_USER_STRING("Show whiskers"));
 
 /* unused
   LOGICAL_DATA(ans)[5] = options->missings_show_p;
@@ -534,17 +547,33 @@ USER_OBJECT_
 RS_GGOBI(setDisplayOptions)(USER_OBJECT_ which, USER_OBJECT_ values,
                               USER_OBJECT_ ggobiId)
 {
-  ggobid *gg = GetGGobi(ggobiId);
+  ggobid *gg;
   gint displayNum, i;
   DisplayOptions *options;
-  displayNum = INTEGER_DATA(which)[0];
+  int apply = 0;
 
-  options = GGOBI(getDisplayOptions)(displayNum, gg);
-  if(options == NULL) {
-    char error_buf[400];
-    sprintf(error_buf,
-	    "No display numbered %d", displayNum);
-    Rf_error(error_buf);
+  if(GET_LENGTH(ggobiId) == 0) {
+	  /* XXX */
+     options = GGOBI(getDefaultDisplayOptions)();
+  } else {
+     gg = GetGGobi(ggobiId);
+     displayNum = INTEGER_DATA(which)[0];
+     options = GGOBI(getDisplayOptions)(displayNum, gg);
+
+     if(options == NULL) {
+	     char error_buf[400];
+	     sprintf(error_buf,
+		     "No display numbered %d", displayNum);
+	     Rf_error(error_buf);
+     }
+     apply = 1;
+  }
+
+
+  if(GET_LENGTH(values) != 8) {
+    PROBLEM "Incorrect length %d for options", 
+	      GET_LENGTH(values)
+    ERROR;
   }
 
   i = 0;
@@ -563,6 +592,12 @@ RS_GGOBI(setDisplayOptions)(USER_OBJECT_ which, USER_OBJECT_ values,
   options->link_p = LOGICAL_DATA(values)[i++];
 */
 
+  if(apply) {
+     displayd *display = GetDisplay(which, ggobiId, NULL);
+     if(display)
+      set_display_options(display, gg);
+  }
+
   return (NULL_USER_OBJECT);
 }
 
@@ -863,7 +898,6 @@ RS_INTERNAL_GGOBI(getPlotDescription)(splotd *splot,
  PROTECT(desc = NEW_LIST(3));
  PROTECT(names = NEW_CHARACTER(3));
 
-/*XX Use S_alloc or R_alloc here. */
  klass = GTK_GGOBI_EXTENDED_SPLOT_CLASS(GTK_OBJECT(splot)->klass);
  if(!klass || !klass->plotted_vars_get) {
    PROBLEM "type of plot (%s) doesn't support the plotted_vars_get method",
@@ -871,6 +905,7 @@ RS_INTERNAL_GGOBI(getPlotDescription)(splotd *splot,
    ERROR;
  }
 
+/*XX Use S_alloc or R_alloc here. */
  vars = (gint *) g_malloc(sizeof(gint)*d->ncols);
  nvars = klass->plotted_vars_get(splot, vars, d);
 
@@ -962,7 +997,7 @@ RS_datasetInstance(datad *d, ggobid *gg)
   USER_OBJECT_ ans, data, name;
   gboolean isEdgeData = (d->edge.n > 0);
 
-  PROTECT(ans = NEW_LIST(2));
+  PROTECT(ans = NEW_LIST(3));
 
 #ifdef USE_EXT_PTR 
   SET_VECTOR_ELT(ans, 0, data = R_MakeExternalPtr((void *) d, Rf_install("GtkGGobiData"), NULL_USER_OBJECT));  
@@ -978,6 +1013,8 @@ RS_datasetInstance(datad *d, ggobid *gg)
 #endif
 
   SET_VECTOR_ELT(ans, 1, RS_ggobiInstance(gg, -1));
+  if(d->name)
+      SET_VECTOR_ELT(ans, 2, mkString(d->name));
 
   PROTECT(name = NEW_CHARACTER(1 + isEdgeData));
   SET_STRING_ELT(name, GET_LENGTH(name)-1, COPY_TO_USER_STRING("ggobiDataset"));
@@ -986,9 +1023,10 @@ RS_datasetInstance(datad *d, ggobid *gg)
   SET_CLASS(ans, name);
 
      /* Put names on the elements of `ans'. */
-  PROTECT(name = NEW_CHARACTER(2));
+  PROTECT(name = NEW_CHARACTER(GET_LENGTH(ans)));
   SET_STRING_ELT(name, 0, COPY_TO_USER_STRING("data"));
   SET_STRING_ELT(name, 1, COPY_TO_USER_STRING("ggobi"));
+  SET_STRING_ELT(name, 2, COPY_TO_USER_STRING("name"));
   SET_NAMES(ans, name);
 
   UNPROTECT(3);
@@ -1046,7 +1084,7 @@ RS_GGOBI(getDescription)(USER_OBJECT_ ggobiId)
 {
  ggobid *gg = GetGGobi(ggobiId);
  datad *d;
- gint numSlots = 3;
+ gint numSlots = 3, numDatasets, i;
  DataMode mode;
  USER_OBJECT_ ans, names, tmp;
  const gchar *tmpname;
@@ -1059,8 +1097,6 @@ RS_GGOBI(getDescription)(USER_OBJECT_ ggobiId)
      return(NULL_USER_OBJECT);
  }
 
- if (g_slist_length (gg->d) == 1)
-   d = (datad *) g_slist_nth_data (gg->d, 0);
 
  PROTECT(ans = NEW_LIST(numSlots));
  PROTECT(names = NEW_CHARACTER(numSlots));
@@ -1077,16 +1113,24 @@ RS_GGOBI(getDescription)(USER_OBJECT_ ggobiId)
  mode = GGOBI(getDataMode)(gg);
  INTEGER_DATA(VECTOR_ELT(ans, 1))[0] = mode;
  PROTECT(tmp = NEW_CHARACTER(1));
+#if 0
+/*XXX */
   SET_STRING_ELT(tmp, 0,
-    COPY_TO_USER_STRING(GGOBI(getDataModeDescription(mode))));
+		 COPY_TO_USER_STRING(GGOBI(getDataModeDescription(mode))));
+#endif
  SET_NAMES(VECTOR_ELT(ans, 1), tmp);
  UNPROTECT(1);
  SET_STRING_ELT(names, 1, COPY_TO_USER_STRING("Data mode"));
 
 
- SET_VECTOR_ELT(ans, 2, NEW_INTEGER(2));
-  INTEGER_DATA(VECTOR_ELT(ans, 2))[0] = d->nrows;
-  INTEGER_DATA(VECTOR_ELT(ans, 2))[1] = d->ncols;
+ numDatasets = g_slist_length (gg->d);
+
+ SET_VECTOR_ELT(ans, 2,  tmp = allocMatrix(INTSXP, numDatasets, 2));
+ for(i = 0; i < numDatasets ; i++) {
+  d = (datad *) g_slist_nth_data (gg->d, i);
+  INTEGER_DATA(tmp)[i] = d->nrows;
+  INTEGER_DATA(tmp)[i + numDatasets] = d->ncols;
+ }
  SET_STRING_ELT(names, 2, COPY_TO_USER_STRING("Data dimensions"));
 
  SET_NAMES(ans, names);
@@ -1243,6 +1287,11 @@ GetGGobi(USER_OBJECT_ id)
    return(ggobi_get(INTEGER_DATA(id)[0]));
  } else if(R_IS(id, "ggobiDataset")) {
     id = VECTOR_ELT(id, 1);
+ } else if(R_IS(id, "GtkGGobi")) {
+    ggobid *gg;
+    gg = (ggobid *) R_ExternalPtrAddr(id);
+    gg = ValidateGGobiRef(gg, false);
+    return(gg);
  }
 
  if(R_IS(id, "ggobi")) {
@@ -1268,6 +1317,18 @@ datad *
 GetDatad(USER_OBJECT_ d)
 {
   ggobid *gg;
+
+  if(R_IS(d, "GtkGGobiData")) {
+    datad *data;
+    data = (datad *) R_ExternalPtrAddr(d);
+    if(data) {
+      if(!ValidateGGobiRef(data->gg, false))
+         return(NULL);
+      data = ValidateDatadRef(data, data->gg, false);
+    }
+    return(data);
+  }
+
   if(!R_IS(d, "ggobiDataset")) {
     return(NULL);
   }
@@ -1276,8 +1337,9 @@ GetDatad(USER_OBJECT_ d)
     datad *data;
     /*  = (datad *) (long) NUMERIC_DATA(VECTOR_ELT(d, 0))[0]; */
     data = (datad *) R_ExternalPtrAddr(VECTOR_ELT(d, 0));
-    if(data)
+    if(data) {
       data = ValidateDatadRef(data, gg, false);
+    }
 
     if(data)
       return(data);
@@ -1306,9 +1368,9 @@ GetDisplay(USER_OBJECT_ rdisplay, USER_OBJECT_ ggobiId, ggobid **setgg)
   } 
 
   if(IS_NUMERIC(rdisplay)) {
-   display  = (displayd*) (long) NUMERIC_DATA(rdisplay)[0];
+     display  = (displayd*) (long) NUMERIC_DATA(rdisplay)[0];
   } else if(TYPEOF(rdisplay) == EXTPTRSXP) {
-      display = (displayd *) R_ExternalPtrAddr(rdisplay);    
+     display = (displayd *) R_ExternalPtrAddr(rdisplay);    
   }
 
   if(display) {
@@ -1391,13 +1453,14 @@ RS_GGOBI(getDisplayTypes)()
   gint n;
   USER_OBJECT_ ans, names;
 
+  l = GGOBI(getExtendedDisplayTypes)();
  
-  n = g_slist_length(ExtendedDisplayTypes); 
+  n = g_slist_length(l); 
   PROTECT(ans = NEW_LIST(n));
   PROTECT(names = NEW_CHARACTER(n));
 
   n = 0;
-  for(l = ExtendedDisplayTypes; l ; l = l->next, n++) {
+  for( ; l ; l = l->next, n++) {
      GtkGGobiExtendedDisplayClass *klass;
      klass = GTK_GGOBI_EXTENDED_DISPLAY_CLASS((GtkObjectClass*) l->data);
      SET_STRING_ELT(names, n, COPY_TO_USER_STRING(klass->titleLabel));
@@ -1461,4 +1524,12 @@ RS_GGOBI(getStructSizes)(USER_OBJECT_ local)
    return(ans);
 }
 
-
+USER_OBJECT_
+RS_GGOBI(getMenubar)(USER_OBJECT_ gobi)
+{
+  ggobid *gg = GetGGobi(gobi);
+  USER_OBJECT_ ans = NULL_USER_OBJECT;
+  if(gg)
+    ans = R_MakeExternalPtr(gg->main_menubar, Rf_install("GtkWidget"), NULL_USER_OBJECT);
+  return(ans);
+}
