@@ -10,8 +10,6 @@
 
 #include "RUtils.h"
 
-USER_OBJECT_ RS_GGOBI(getBrushColor)(USER_OBJECT_ ggobiId);
-
 
 /*
   Returns the names of the datasets within the specified
@@ -73,6 +71,33 @@ RS_INTERNAL_GGOBI(getVariableNames)(datad *d)
 }
 
 
+USER_OBJECT_
+createFactor(USER_OBJECT_ vals, vartabled *vt, datad *d, int which)
+{
+  USER_OBJECT_ labels, levels, ans, e;
+  int i;
+
+  PROTECT(levels = NEW_INTEGER(vt->nlevels));
+  PROTECT(labels = NEW_CHARACTER(vt->nlevels));
+  for(i = 0; i < vt->nlevels; i++) {
+     INTEGER_DATA(levels)[i] = vt->level_values[i];
+     SET_STRING_ELT(labels, i, COPY_TO_USER_STRING(vt->level_names[i]));
+  }
+
+  PROTECT(e = allocVector(LANGSXP, 4));
+  SETCAR(e, Rf_install("factor"));
+  SETCAR(CDR(e), vals);
+  SETCAR(CDR(CDR(e)), levels);
+  SETCAR(CDR(CDR(CDR(e))), labels);
+
+  ans = eval(e, R_GlobalEnv);
+
+  UNPROTECT(3);
+ 
+  return(ans);
+}
+
+#if 0
 USER_OBJECT_
 RS_GGOBI(getData)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 {
@@ -108,7 +133,7 @@ RS_GGOBI(getData)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
           vt = vartable_element_get(j, d);
    	  SET_STRING_ELT(colnames, j, COPY_TO_USER_STRING(vt->collab));
    	}
-        if(d->missing.vals[i][j])
+        if(d->missing.vals && d->missing.vals[i][j])
   	   NUMERIC_DATA(ans)[i + j*nr] = NA_REAL;
         else
       	   NUMERIC_DATA(ans)[i + j*nr] = d->raw.vals[i][j];
@@ -123,6 +148,103 @@ RS_GGOBI(getData)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 
  return (ans);
 }
+#else
+USER_OBJECT_
+RS_GGOBI(getData)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+ datad *d;
+ USER_OBJECT_ ans, colnames, el;
+ gint nr, nc;
+ gint i, j, m;
+ vartabled *vt;
+
+ ans = NULL_USER_OBJECT;
+
+ d = resolveDatad(datasetId, ggobiId, NULL);
+
+ if(d) {
+   nr = d->nrows_in_plot;
+   nc = d->ncols;
+ 
+   if(nr == 0 || nc == 0)
+       return(NULL_USER_OBJECT);
+
+#if 0
+   PROTECT(rownames = NEW_CHARACTER(nr));
+#endif
+   PROTECT(colnames = NEW_CHARACTER(nc));
+
+   PROTECT(ans = NEW_LIST(nc));
+
+   for(j = 0; j < nc; j++) {
+      vt = vartable_element_get(j, d);
+      SET_STRING_ELT(colnames, j, COPY_TO_USER_STRING(vt->collab));
+
+      PROTECT(el= NEW_NUMERIC(nr));
+      for (m = 0; m < nr; m++) {
+	i = d->rows_in_plot.els[m];
+        if(d->missing.vals && d->missing.vals[i][j])
+  	   NUMERIC_DATA(el)[m] = NA_REAL;
+        else
+      	   NUMERIC_DATA(el)[m] = d->raw.vals[i][j];
+      }
+      if(vt->vartype == categorical) {
+        PROTECT(el = createFactor(el, vt, d, j));
+      }
+      SET_VECTOR_ELT(ans, j, el);
+      UNPROTECT(1 + (vt->vartype == categorical));
+   }
+
+#if 0
+   for (m =0; m < nr; m++) {
+     i = d->rows_in_plot.els[m];
+      SET_STRING_ELT(rownames, m, 
+   	COPY_TO_USER_STRING ((gchar *) g_array_index (d->rowlab, gchar *, i)));
+
+   }
+   SET_ATTRIB(ans, R_RowNamesSymbol, rownames);
+
+#endif
+   SET_NAMES(ans, colnames);
+
+   UNPROTECT(2);
+ }
+   return(ans);
+}
+#endif
+
+USER_OBJECT_
+RS_GGOBI(getDatasetRecordIds)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+ datad *d;
+ USER_OBJECT_ ans;
+ int i, n, m;
+
+ d = resolveDatad(datasetId, ggobiId, NULL);
+
+ if(!d) {
+  PROBLEM "No such dataset"
+  ERROR;
+ }
+
+ if(!d->rowIds) {
+    return(NULL_USER_OBJECT);
+ }
+
+
+ n = d->nrows_in_plot;
+
+ PROTECT(ans = NEW_CHARACTER(n));
+ for(m = 0; m < n ; m++) {
+   i = d->rows_in_plot.els[m];
+    if(d->rowIds[i])
+     SET_STRING_ELT(ans, m, COPY_TO_USER_STRING(d->rowIds[i]));
+ }
+
+ UNPROTECT(1);
+
+ return(ans);
+}
 
 
 
@@ -136,7 +258,7 @@ RS_GGOBI(getSelectedIndices)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 
   d = resolveDatad(datasetId, ggobiId, NULL);
   if(d) {
-   gint nr, i, ctr;
+    gint nr, i, m, ctr;
    gchar *name;
        nr = d->npts_under_brush;
        
@@ -146,7 +268,8 @@ RS_GGOBI(getSelectedIndices)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
        PROTECT(ans = NEW_INTEGER(nr));
        PROTECT(names = NEW_CHARACTER(nr));
        ctr = 0;
-       for(i = 0 ; i < d->nrows ; i++) {
+       for(m = 0 ; m < d->nrows ; m++) {
+         i = d->rows_in_plot.els[m];
       	 if (d->pts_under_brush.els[i]) {
            INTEGER_DATA(ans)[ctr] = i + 1; /* Make these 1 based */
            name = (gchar *) g_array_index (d->rowlab, gchar *, i);
@@ -168,9 +291,9 @@ USER_OBJECT_
 RS_GGOBI(getRowNames)(USER_OBJECT_ data, USER_OBJECT_ ggobiId)
 {
   USER_OBJECT_ ans = NULL_USER_OBJECT;
-  gint nr, i;
+  gint nr, i, m;
   datad *d;
-  if(R_IS(data, "ggobiDataset")) {
+  if(R_IS(data, "ggobiDataset") || R_IS(data, "GtkGGobiData")) {
     d = GetDatad(data);
   } else if(IS_INTEGER(data)) {
     ggobid *gg = GetGGobi(ggobiId);
@@ -180,11 +303,12 @@ RS_GGOBI(getRowNames)(USER_OBJECT_ data, USER_OBJECT_ ggobiId)
   }
 
   if(d) {
-    nr = d->nrows;
+    nr = d->nrows_in_plot;
     PROTECT(ans = NEW_CHARACTER(nr));
 
-    for (i = 0 ; i < nr ; i++) {
-      SET_STRING_ELT(ans, i,
+    for (m = 0 ; m < nr ; m++) {
+      i = d->rows_in_plot.els[m];
+      SET_STRING_ELT(ans, m,
       COPY_TO_USER_STRING ((gchar *) g_array_index (d->rowlab, gchar *, i)));
     }
     UNPROTECT(1);
@@ -199,16 +323,17 @@ RS_GGOBI(getRowColors)(USER_OBJECT_ ggobiId)
   ggobid *gg = GetGGobi(ggobiId);
   datad *d;
   USER_OBJECT_ ans;
-  gint nr, i;
+  gint nr, i, m;
 
   if (g_slist_length (gg->d) == 1)
     d = (datad *) g_slist_nth_data (gg->d, 0);
-  nr = d->nrows;
+  nr = d->nrows_in_plot;
 
   PROTECT(ans = NEW_INTEGER(nr));
 
-  for (i = 0; i < nr ; i++) {
-    INTEGER_DATA(ans)[i] = d->color_now.els[i];
+  for (m = 0; m < nr ; m++) {
+    i = d->rows_in_plot.els[i];
+    INTEGER_DATA(ans)[m] = d->color_now.els[i];
   }
 
   UNPROTECT(1);
@@ -241,24 +366,70 @@ RS_GGOBI(getSelectedVariables)(USER_OBJECT_ ggobiId)
 
 
 USER_OBJECT_
-RS_GGOBI(getConnectedEdges)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+RS_GGOBI(getSymbolicEdges)(USER_OBJECT_ edgesetId, USER_OBJECT_ ggobiId)
 {
- datad *d = resolveDatad(datasetId, ggobiId, NULL);
- gint i, ctr; 
+ datad *e = resolveDatad(edgesetId, ggobiId, NULL);
  USER_OBJECT_ ans, dim;
- gint n = d->edge.n;
+ gint i, ctr, n;
 
- ans = NEW_INTEGER(n * 2);
- PROTECT(ans);
+ n = e->edge.n;
+
+ if(!e) {
+   PROBLEM "Invalid ggobi dataset identifier(s)"
+   ERROR;
+ }
+
+ PROTECT(ans = NEW_CHARACTER(n * 2));
 
  ctr = 0;
-#ifdef OLD_STYLE_IDS
  for(i = 0; i < n; i++) {
-   INTEGER_DATA(ans)[ctr] = d->edge.old_endpoints[i].a; 
-   INTEGER_DATA(ans)[ctr+n] = d->edge.old_endpoints[i].b; 
+   SET_STRING_ELT(ans, ctr, COPY_TO_USER_STRING(e->edge.sym_endpoints[i].a));
+   SET_STRING_ELT(ans, ctr+n, COPY_TO_USER_STRING(e->edge.sym_endpoints[i].b));
+   ctr++;
+ }
+
+ /* Now set the dim() on this vector to make it a matrix. 
+    The S code will put the dimnames on it.
+  */
+ PROTECT(dim = NEW_INTEGER(2));
+  INTEGER_DATA(dim)[0] = n;
+  INTEGER_DATA(dim)[1] = 2;
+  SET_DIM(ans, dim);
+ UNPROTECT(2);
+
+ return(ans); 
+}
+
+
+USER_OBJECT_
+RS_GGOBI(getConnectedEdges)(USER_OBJECT_ edgesetId, USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+ datad *d = resolveDatad(datasetId, ggobiId, NULL);
+ datad *e = resolveDatad(edgesetId, ggobiId, NULL);
+ gint ctr; 
+ USER_OBJECT_ ans, dim;
+ gint i, n;
+ endpointsd *endpoints;
+
+ if(!d || !e) {
+   PROBLEM "Invalid ggobi dataset identifier(s)"
+   ERROR;
+ }
+
+ n = e->edge.n;
+
+ endpoints = resolveEdgePoints(e, d);
+ if(!endpoints) 
+    return(NULL_USER_OBJECT);
+
+ PROTECT(ans = NEW_INTEGER(n * 2));
+
+ ctr = 0;
+ for(i = 0; i < n; i++) {
+   INTEGER_DATA(ans)[ctr] = endpoints[i].a;  
+   INTEGER_DATA(ans)[ctr+n] = endpoints[i].b; 
    ctr++;
  }
-#endif
 
  /* Now set the dim() on this vector to make it a matrix. 
     The S code will put the dimnames on it.
@@ -274,9 +445,8 @@ RS_GGOBI(getConnectedEdges)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 
 
 USER_OBJECT_
-RS_GGOBI(getRgroups)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+RS_GGOBI(getRowsInPlot)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 {
-
   USER_OBJECT_ ans = NULL_USER_OBJECT;
   datad *d;
  
@@ -298,94 +468,83 @@ RS_GGOBI(getRgroups)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 }
 
 
-
 USER_OBJECT_
-RS_GGOBI(setRgroups)(USER_OBJECT_ rowIds, USER_OBJECT_ datasetId, USER_OBJECT_ gobiId)
+RS_GGOBI(getDataAttribute)(vector_b v)
 {
-  gint i;
+  USER_OBJECT_ ans = NULL_USER_OBJECT;
   datad *d;
-  USER_OBJECT_ ans = NEW_LOGICAL(1);
-
-  d = resolveDatad(datasetId, gobiId, NULL);
+ 
   if(d) {
-    int num = GET_LENGTH(rowIds);
-       /* Why do we free this? Can't we just reuse it, or does it have
-          the wrong length? Weird if it does, as we don't seem to reset 
-          the length. 
-        */
-/*
- *  if (d->rows_in_plot)
- *    g_free(d->rows_in_plot);
- *  d->rows_in_plot = (gint *) g_malloc(sizeof(gint)*d->nrows_in_plot);
-*/
-    /* dfs: not sure what the intention is in this function ... */
-    vectori_realloc (&d->rows_in_plot, d->nrows);
+    gint i;
+    PROTECT(ans = NEW_LOGICAL(v.nels));
 
-    for (i = 0; i < num ; i++) {
-      d->rows_in_plot.els[i] = INTEGER_DATA(rowIds)[i];
+    for(i = 0; i < v.nels; i++) {
+       LOGICAL_DATA(ans)[i] = v.els[i];
     }
-    LOGICAL_DATA(ans)[0] = TRUE;
+
+    UNPROTECT(1);
   }
 
-  return(ans);
+  return (ans);
 }
 
 
-
-USER_OBJECT_
-RS_GGOBI(getColorMap)(USER_OBJECT_ ggobiId)
+void
+RS_GGOBI(setDataAttribute)(vector_b *v, USER_OBJECT_ vals, datad *d)
 {
-  USER_OBJECT_ ans;
-  static const gchar *const ColorNames[]={"Red", "Green", "Blue"};
-  ggobid *gg = GetGGobi(ggobiId);
-  const gchar *colorName;
-  USER_OBJECT_ cnames, rnames, dm;
-  colorschemed *scheme = gg->activeColorScheme;
-  gint nr = scheme->n, i;
-
-#ifdef _R_
-  PROTECT(ans = allocMatrix(REALSXP, nr, 3));
-#else
-#error "Add code to allocate a matrix here"
-#endif
-
-  PROTECT(rnames = NEW_CHARACTER(nr));
-  for(i = 0; i < nr; i++) {
-    NUMERIC_DATA(ans)[i] = scheme->rgb[i].red;
-    NUMERIC_DATA(ans)[nr + i] = scheme->rgb[i].green;
-    NUMERIC_DATA(ans)[2*nr + i] = scheme->rgb[i].blue;
-    colorName = GGOBI(getColorName)(i, gg, true);
-    if(colorName && colorName[0])
-      SET_STRING_ELT(rnames, i, COPY_TO_USER_STRING(colorName));
+  USER_OBJECT_ ans = NULL_USER_OBJECT;
+ 
+  if(v) {
+    gint i, n;
+    n = GET_LENGTH(vals);
+    if(n != d->nrows) {
+	 PROBLEM "number of values must be the same as the number of records in the GGobi dataset"
+	 ERROR;
+    }
+    vectorb_realloc(v, n);
+    for(i = 0; i < n; i++) {
+       v->els[i] = LOGICAL_DATA(ans)[i];
+    }
   }
-
-  PROTECT(cnames = NEW_CHARACTER(3));
-  for(i = 0; i < 3; i++)
-   SET_STRING_ELT(cnames, i, COPY_TO_USER_STRING(ColorNames[i]));
-
-  PROTECT(dm = NEW_LIST(2));
-  SET_VECTOR_ELT(dm, 0, rnames);
-  SET_VECTOR_ELT(dm, 1, cnames);
-  SET_DIMNAMES(ans, dm);
-  UNPROTECT(4);
-
- return(ans);
 }
 
 
 USER_OBJECT_
-RS_GGOBI(setColorMap)(USER_OBJECT_ vals, USER_OBJECT_ names, USER_OBJECT_ ggobiId)
+RS_GGOBI(getSampledIndices)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 {
-  ggobid *gg = GetGGobi(ggobiId);
+  datad *d = resolveDatad(datasetId, ggobiId, NULL);
+  return(RS_GGOBI(getDataAttribute)(d->sampled));
+}
 
-  GGOBI(setColorMap)(NUMERIC_DATA(vals), GET_LENGTH(vals)/3, gg);
-  if(GET_LENGTH(names)) {
-          
-  }
-  
+USER_OBJECT_
+RS_GGOBI(setSampledIndices)(USER_OBJECT_ vals, USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+  datad *d = resolveDatad(datasetId, ggobiId, NULL);
+  RS_GGOBI(setDataAttribute)(&(d->sampled), vals, d);
+  rows_in_plot_set(d, d->gg);
   return(NULL_USER_OBJECT);
 }
 
+USER_OBJECT_
+RS_GGOBI(getExcludedIndices)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+  datad *d = resolveDatad(datasetId, ggobiId, NULL);
+  return(RS_GGOBI(getDataAttribute)(d->excluded));
+}
+
+USER_OBJECT_
+RS_GGOBI(setExcludedIndices)(USER_OBJECT_ vals, USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+  datad *d = resolveDatad(datasetId, ggobiId, NULL);
+  RS_GGOBI(setDataAttribute)(&(d->excluded), vals, d);
+  rows_in_plot_set(d, d->gg);
+  return(NULL_USER_OBJECT);
+}
+
+
+
+
+
 USER_OBJECT_
 RS_GGOBI(setMode)(USER_OBJECT_ name, USER_OBJECT_ ggobiId)
 {
@@ -525,17 +684,17 @@ RS_GGOBI(setCasesHidden)(USER_OBJECT_ vals, USER_OBJECT_ which, USER_OBJECT_ dat
 USER_OBJECT_
 RS_GGOBI(getCasesHidden)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 {
-  gint num, i;
   USER_OBJECT_ ans = NULL_USER_OBJECT;
   ggobid *gg;
   datad *d;
   d = resolveDatad(datasetId, ggobiId,  &gg);
     if(d) {
-      num = d->nrows, i;
+      gint num = d->nrows_in_plot, i, m;
 
       PROTECT(ans = NEW_LOGICAL(num));
-      for(i = 0; i < num; i++) {
-        LOGICAL_DATA(ans)[i] = GGOBI(getCaseHidden)(i, d, gg);
+      for(m = 0; m < num; m++) {
+        i = d->rows_in_plot.els[m];
+        LOGICAL_DATA(ans)[m] = GGOBI(getCaseHidden)(i, d, gg);
       }
 
       UNPROTECT(1);
@@ -547,23 +706,95 @@ RS_GGOBI(getCasesHidden)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 
 
 
+/*
+ This allows us to add a variable to the existing dataset.
+ If there are no variables currently in the dataset, the first
+ one added will induce a call to GGOBI(setData) and
+ datad_init() which will create the relevant tab in the
+ notebook. However, it will also slight mess up with the
+ addition of the variables into the control panel
+ (e.g. create X, Y and Z buttons for no good reason.)
+
+ We might want to inhibit this and wait until the end.
+*/
 USER_OBJECT_
-RS_GGOBI(addVariable)(USER_OBJECT_ vals, USER_OBJECT_ name, USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+RS_GGOBI(addVariable)(USER_OBJECT_ vals, USER_OBJECT_ name, USER_OBJECT_ levels,
+		      USER_OBJECT_ values, 
+                      USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
 {
   datad *d;
   ggobid *gg;
   USER_OBJECT_ ans = NULL_USER_OBJECT;
   d = resolveDatad(datasetId, ggobiId, &gg);
 
+
   if(d) {
-    ans = NEW_INTEGER(1);
-    INTEGER_DATA(ans)[0] = GGOBI(addVariable)(NUMERIC_DATA(vals),
-      GET_LENGTH(vals), CHAR_DEREF(STRING_ELT(name,0)), true, d, gg);
+    int i;
+    PROTECT(ans = NEW_INTEGER(1));
+    if(GET_LENGTH(levels)) {
+     USER_OBJECT_ names;
+     int n = GET_LENGTH(levels);
+     char **levelValues;
+
+     names = GET_NAMES(levels);
+     levelValues = (char **) S_alloc(n, sizeof(char *));
+     for(i = 0; i < n ; i++) {
+         levelValues[i] = CHAR_DEREF(STRING_ELT(names, i));
+     }
+
+     INTEGER_DATA(ans)[0] = GGOBI(addCategoricalVariable)(NUMERIC_DATA(vals),
+							      GET_LENGTH(vals), 
+							      CHAR_DEREF(STRING_ELT(name,0)), 
+							      levelValues, 
+ 	 						      INTEGER_DATA(values), INTEGER_DATA(levels),
+							      n,
+							      true, d, gg);
+    } else {
+         INTEGER_DATA(ans)[0] = GGOBI(addVariable)(NUMERIC_DATA(vals),
+					       GET_LENGTH(vals), 
+					       CHAR_DEREF(STRING_ELT(name,0)), 
+					       true, d, gg);
+    }
+    UNPROTECT(1);
+  } else {
+    PROBLEM "Cannot resolve dataset in GGobi"
+    ERROR;
   }
 
   return(ans);
 }
 
+USER_OBJECT_
+RS_GGOBI(datad_init)(USER_OBJECT_ cleanup, USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+  datad *d;
+  ggobid *gg;
+
+  d = resolveDatad(datasetId, ggobiId, &gg);
+
+  if(d) {
+     datad_init(d, gg, LOGICAL_DATA(cleanup)[0]);
+  }
+
+  return(NULL_USER_OBJECT);
+}
+
+USER_OBJECT_
+RS_GGOBI(varpanel_populate)(USER_OBJECT_ datasetId, USER_OBJECT_ ggobiId)
+{
+  datad *d;
+  ggobid *gg;
+
+  d = resolveDatad(datasetId, ggobiId, &gg);
+
+  if(d) {
+     varpanel_populate(d, gg);
+  }
+
+  return(NULL_USER_OBJECT);
+}
+
+
 USER_OBJECT_
 RS_GGOBI(setVariableValues)(USER_OBJECT_ vals, USER_OBJECT_ rowIds,
                              USER_OBJECT_ colId, USER_OBJECT_ update, 
@@ -693,7 +924,8 @@ RS_GGOBI(raiseDisplay)(glong *plotId, glong *numEls, glong *raiseOrIcon,
 USER_OBJECT_
 RS_GGOBI(getVariable)(int which, datad *d)
 {
-    int nr = d->nrows, i;
+  int nr = d->nrows_in_plot, i, m;
+  float f;
     USER_OBJECT_ ans;
     vartabled *vt;
      vt = vartable_element_get (which, d);
@@ -703,21 +935,23 @@ RS_GGOBI(getVariable)(int which, datad *d)
      } else {
        PROTECT(ans = NEW_NUMERIC(nr));
      }
-    for(i = 0; i < nr; i++) {
-        float f = d->raw.vals[i][which];
-    if(vt->vartype == categorical)
-	    INTEGER_DATA(ans)[i] = f;
-        else
-	    NUMERIC_DATA(ans)[i] = f;
+    for(m = 0; m < nr; m++) {
+      i = d->rows_in_plot.els[m];
+      f = d->raw.vals[i][which];
+      if(vt->vartype == categorical)
+	    INTEGER_DATA(ans)[m] = f;
+      else
+	    NUMERIC_DATA(ans)[m] = f;
     }
     if(vt->vartype == categorical) {
         USER_OBJECT_ levs;
         nr = vt->nlevels;
         PROTECT(levs = NEW_CHARACTER(nr));
-	for(i = 0; i < nr; i++) {
+	for(m = 0; m < nr; m++) {
           gchar* tmp;
+          i = d->rows_in_plot.els[m];
           tmp = (gchar *) vt->level_names[i];
-	  SET_STRING_ELT(levs, i, COPY_TO_USER_STRING(tmp));
+	  SET_STRING_ELT(levs, m, COPY_TO_USER_STRING(tmp));
 	}
         SET_LEVELS(ans, levs);
         PROTECT(levs = NEW_CHARACTER(1));
@@ -757,4 +991,95 @@ RS_GGOBI(getVariables)(USER_OBJECT_ which, USER_OBJECT_ datasetId, USER_OBJECT_
 }
 
 
+USER_OBJECT_
+RS_GGOBI(getMainWindow)(USER_OBJECT_ gobiId)
+{
+  ggobid *gg = GetGGobi(gobiId);
+  USER_OBJECT_ ans;
+
+  ans = R_MakeExternalPtr(gg->main_window, Rf_install("GtkWindow"), NULL_USER_OBJECT);
+
+  return(ans);
+}
+
+USER_OBJECT_
+RS_GGOBI(getCurrentDisplay)(USER_OBJECT_ gobiId)
+{
+  ggobid *gg = GetGGobi(gobiId);
+  USER_OBJECT_ ans;
+
+  ans = R_MakeExternalPtr(gg->current_display, Rf_install("GtkWidget"), NULL_USER_OBJECT);
+
+  return(ans);
+}
+
+
+
+#include "plugin.h"
+
+USER_OBJECT_
+RS_GGOBI(getPlugins)(USER_OBJECT_ gobiId)
+{
+
+  ggobid *gg = GetGGobi(gobiId);
+  USER_OBJECT_ ans, names;
+  GList *tmp;
+  gint n, i;
+
+  n = g_list_length(gg->pluginInstances);
+
+  PROTECT(ans = NEW_LIST(n));
+  PROTECT(names = NEW_CHARACTER(n));
+  for(i = 0, tmp = gg->pluginInstances; i < n; tmp=tmp->next, i++) {
+     PluginInstance *inst = tmp->data;
+     SET_VECTOR_ELT(ans, i, R_MakeExternalPtr(tmp->data, Rf_install("GGobiPluginInstance"), NULL_USER_OBJECT));
+     SET_STRING_ELT(names, i, COPY_TO_USER_STRING(inst->info->details->name));
+  }
+  SET_NAMES(ans, names);
+  UNPROTECT(2);
+
+  return(ans);
+}
+
+
+static USER_OBJECT_
+R_getPluginInfo(GGobiPluginInfo *info)
+{
+  SEXP klass = MAKE_CLASS("GGobiPluginInfo");
+  USER_OBJECT_ ans;
+
+  PROTECT(ans = NEW(klass));
+  SET_SLOT(ans, Rf_install("dll"), mkString(info->details->dllName));
+  SET_SLOT(ans, Rf_install("name"), mkString(info->details->name));
+  SET_SLOT(ans, Rf_install("description"), mkString(info->details->description));
+  SET_SLOT(ans, Rf_install("author"), mkString(info->details->author));
+
+  UNPROTECT(1);
+
+  return(ans);
+}
+
+
+USER_OBJECT_
+RS_GGOBI(getPluginInfo)(USER_OBJECT_ gobiId)
+{
+  USER_OBJECT_ ans, names;
+  GList *tmp, *plugins;
+  gint n, i;
+  plugins = GGOBI_getSessionOptions()->info->plugins;
+  n = g_list_length(plugins);
+
+  PROTECT(ans = NEW_LIST(n));
+  PROTECT(names = NEW_CHARACTER(n));
+  for(i = 0, tmp = plugins; i < n; tmp=tmp->next, i++) {
+     GGobiPluginInfo *inst = tmp->data;
+     SET_VECTOR_ELT(ans, i, R_getPluginInfo(inst));
+     SET_STRING_ELT(names, i, COPY_TO_USER_STRING(inst->details->name));
+  }
+  SET_NAMES(ans, names);
+  UNPROTECT(2);
+
+  return(ans);
+}
+
 
