@@ -22,6 +22,19 @@ PrintValue(obj);
 }  
 
 
+SEXP
+isSameRef(USER_OBJECT_ x, USER_OBJECT_ y)
+{
+   SEXP ans;
+   void *a, *b;
+   
+   ans = NEW_LOGICAL(1);
+   a = R_ExternalPtrAddr(x);
+   b = R_ExternalPtrAddr(y);
+   INTEGER_DATA(ans)[0] = (a == b);
+   return(ans);
+}
+
 
 void
 RS_throwError(char *msg)
